@@ -18,12 +18,15 @@ import io
 from enum import Enum, unique
 import json
 
+from typing import List
+
 from . import rest
 
 PATH_ENGINE = "/compute"
 PATH_DATABASE = "/database"
 PATH_TRANSACTION = "/transaction"
 PATH_USER = "/users"
+PATH_OAUTH_CLIENT = "/oauth-clients"
 
 
 # Engine sizes
@@ -47,26 +50,77 @@ class Mode(str, Enum):
     CLONE_OVERWRITE = "CLONE_OVERWRITE"
 
 
+# User roles
+@unique
+class Role(str, Enum):
+    USER = "user"
+    ADMIN = "admin"
+
+
+# User/OAuth-client permissions
+@unique
+class Permission(str, Enum):
+    # computes
+    CREATE_COMPUTE = "create:compute"
+    DELETE_COMPUTE = "delete:compute"
+    LIST_COMPUTES = "list:compute"
+    READ_COMPUTE = "read:compute"
+    # databases
+    LIST_DATABASES = "list:database"
+    UPDATE_DATABASE = "update:database"
+    DELETE_DATABASE = "delete:database"
+    # transactions
+    RUN_TRANSACTION = "run:transaction"
+    # credits
+    READ_CREDITS_USAGE = "read:credits_usage"
+    # oauth clients
+    CREATE_OAUTH_CLIENT = "create:oauth_client"
+    READ_OAUTH_CLIENT = "read:oauth_client"
+    LIST_OAUTH_CLIENTS = "list:oauth_client"
+    UPDATE_OAUTH_CLIENT = "update:oauth_client"
+    DELETE_OAUTH_CLIENT = "delete:oauth_client"
+    ROTATE_OAUTH_CLIENT_SECRET = "rotate:oauth_client_secret"
+    # users
+    CREATE_USER = "create:user"
+    LIST_USERS = "list:user"
+    READ_USER = "read:user"
+    UPDATE_USER = "update:user"
+    # roles
+    LIST_ROLES = "list:role"
+    READ_ROLE = "read:role"
+    # permissions
+    LIST_PERMISSIONS = "list:permission"
+    # access keys
+    CREATE_ACCESS_KEY = "create:accesskey"
+    LIST_ACCESS_KEYS = "list:accesskey"
+
+
 __all__ = [
     "Context",
     "EngineSize",
     "Mode",
+    "Role",
+    "Permission",
     "create_database",
     "create_engine",
     "create_user",
+    "create_oauth_client",
     "delete_database",
     "delete_engine",
-    "delete_source",
-    "delete_user",
+    "delete_model",
+    "disable_user",
+    "delete_oauth_client",
     "get_database",
     "get_engine",
-    "get_source",
+    "get_model",
     "get_user",
+    "get_oauth_client",
     "list_databases",
-    "list_edb",
+    "list_edbs",
     "list_engines",
-    "list_sources",
+    "list_models",
     "list_users",
+    "list_oauth_clients",
     "load_csv",
     "query",
 ]
@@ -117,13 +171,29 @@ def create_engine(ctx: Context, engine: str, size: EngineSize = EngineSize.XS):
     return json.loads(rsp)
 
 
-def create_user(ctx: Context, user: str):
-    raise Exception("not implemented")
+def create_user(ctx: Context, user: str, roles: List[Role] = None):
+    rs = roles or [Role.USER]
+    data = {
+        "email": user,
+        "roles": [r.value for r in rs]}
+    url = _mkurl(ctx, PATH_USER)
+    rsp = rest.post(ctx, url, data)
+    return json.loads(rsp)
+
+
+def create_oauth_client(ctx: Context, name: str, permissions: List[Permission] = None):
+    ps = permissions or []
+    data = {
+        "name": name,
+        "permissions": ps}
+    url = _mkurl(ctx, PATH_OAUTH_CLIENT)
+    rsp = rest.post(ctx, url, data)
+    return json.loads(rsp)["client"]
 
 
 # Derives the database open_mode based on the given arguments.
-def _create_mode(source_name: str, overwrite: bool) -> Mode:
-    if source_name is not None:
+def _create_mode(source_database: str, overwrite: bool) -> Mode:
+    if source_database is not None:
         result = Mode.CLONE_OVERWRITE if overwrite else Mode.CLONE
     else:
         result = Mode.CREATE_OVERWRITE if overwrite else Mode.CREATE
@@ -138,17 +208,27 @@ def delete_engine(ctx: Context, engine: str) -> dict:
 
 
 def delete_database(ctx: Context, database: str) -> dict:
-    url = _mkurl(ctx, f"{PATH_DATABASE}/{database}")
+    data = {"name": database}
+    url = _mkurl(ctx, PATH_DATABASE)
+    rsp = rest.delete(ctx, url, data)
+    return json.loads(rsp)
+
+
+def disable_user(ctx: Context, user: str) -> dict:
+    data = {"status": "INACTIVE"}
+    url = _mkurl(ctx, f"{PATH_USER}/{user}")
+    rsp = rest.patch(ctx, url, data)
+    return json.loads(rsp)
+
+
+def delete_oauth_client(ctx: Context, id: str) -> dict:
+    url = _mkurl(ctx, f"{PATH_OAUTH_CLIENT}/{id}")
     rsp = rest.delete(ctx, url, None)
     return json.loads(rsp)
 
 
-def delete_user(ctx: Context, user: str) -> dict:
-    raise Exception("not implemented")
-
-
 def get_engine(ctx: Context, engine: str) -> dict:
-    return _get_resource(ctx, PATH_ENGINE, name=engine, key="computes")
+    return _get_resource(ctx, PATH_ENGINE, name=engine, deleted_on="", key="computes")
 
 
 def get_database(ctx: Context, database: str) -> dict:
@@ -157,7 +237,10 @@ def get_database(ctx: Context, database: str) -> dict:
 
 def get_user(ctx: Context, user: str) -> dict:
     return _get_resource(ctx, f"{PATH_USER}/{user}", name=user)
-    # return _get_resource(ctx, PATH_USER, name=user, key="users")
+
+
+def get_oauth_client(ctx: Context, id: str) -> dict:
+    return _get_resource(ctx, f"{PATH_OAUTH_CLIENT}/{id}", key="client")
 
 
 def list_engines(ctx: Context, state=None) -> list:
@@ -178,6 +261,10 @@ def list_users(ctx: Context) -> list:
     return _list_collection(ctx, PATH_USER, key="users")
 
 
+def list_oauth_clients(ctx: Context) -> list:
+    return _list_collection(ctx, PATH_OAUTH_CLIENT, key="clients")
+
+
 #
 # Transaction endpoint
 #
@@ -212,7 +299,7 @@ class Transaction(object):
             "nowait_durable": self.nowait_durable,
             "readonly": self.readonly,
             "type": "Transaction",
-            "version": 0  # 25
+            "version": 0
         }
         if self.engine is not None:
             result["computeName"] = self.engine
@@ -236,12 +323,12 @@ class Transaction(object):
         return json.loads(rsp)
 
 
-def _delete_source_action(name: str) -> dict:
+def _delete_model_action(name: str) -> dict:
     return {"type": "ModifyWorkspaceAction", "delete_source": [name]}
 
 
-def _install_source_action(name: str, source: str) -> dict:
-    return {"type": "InstallAction", "sources": [_source(name, source)]}
+def _install_model_action(name: str, model: str) -> dict:
+    return {"type": "InstallAction", "sources": [_model(name, model)]}
 
 
 def _list_action():
@@ -277,34 +364,34 @@ def _query_action_input(name: str, value) -> dict:
 
 
 # `inputs`: map of parameter name to input value
-def _query_action(source: str, inputs: dict = None, outputs: list = None) -> dict:
+def _query_action(model: str, inputs: dict = None, outputs: list = None) -> dict:
     inputs = inputs or {}
     inputs = [_query_action_input(k, v) for k, v in inputs.items()]
     return {
         "type": "QueryAction",
-        "source": _source("query", source),
+        "source": _model("query", model),
         "persist": [],
         "inputs": inputs,
         "outputs": outputs or []}
 
 
-def _source(name: str, source: str) -> dict:
+def _model(name: str, model: str) -> dict:
     return {
         "type": "Source",
         "name": name,
         "path": "",       # todo: check if required?
-        "value": source}
+        "value": model}
 
 
-# Returns full list of source objects, including source values.
-def _list_sources(ctx: Context, database: str, engine: str) -> dict:
+# Returns full list of models.
+def _list_models(ctx: Context, database: str, engine: str) -> dict:
     tx = Transaction(database, engine, mode=Mode.OPEN)
     rsp = tx.run(ctx, _list_action())
     actions = rsp["actions"]
     assert len(actions) == 1
     action = actions[0]
-    sources = action["result"]["sources"]
-    return sources
+    models = action["result"]["sources"]
+    return models
 
 
 def create_database(ctx: Context, database: str, engine: str,
@@ -314,29 +401,29 @@ def create_database(ctx: Context, database: str, engine: str,
     return tx.run(ctx)
 
 
-def delete_source(ctx: Context, database: str, engine: str, source: str) -> dict:
+def delete_model(ctx: Context, database: str, engine: str, model: str) -> dict:
     tx = Transaction(database, engine, mode=Mode.OPEN, readonly=False)
-    actions = [_delete_source_action(source)]
+    actions = [_delete_model_action(model)]
     return tx.run(ctx, *actions)
 
 
-# Returns the source value for the named source.
-def get_source(ctx: Context, database: str, engine: str, name: str) -> str:
-    sources = _list_sources(ctx, database, engine)
-    for source in sources:
-        if source["name"] == name:
-            return source["value"]
-    raise Exception(f"source '{name}' not found")
+# Returns the named model
+def get_model(ctx: Context, database: str, engine: str, name: str) -> str:
+    models = _list_models(ctx, database, engine)
+    for model in models:
+        if model["name"] == name:
+            return model["value"]
+    raise Exception(f"model '{name}' not found")
 
 
-def install_source(ctx: Context, database: str, engine: str, sources: dict) -> dict:
+def install_model(ctx: Context, database: str, engine: str, models: dict) -> dict:
     tx = Transaction(database, engine, mode=Mode.OPEN, readonly=False)
-    actions = [_install_source_action(name, source)
-               for name, source in sources.items()]
+    actions = [_install_model_action(name, model)
+               for name, model in models.items()]
     return tx.run(ctx, *actions)
 
 
-def list_edb(ctx: Context, database: str, engine: str) -> list:
+def list_edbs(ctx: Context, database: str, engine: str) -> list:
     tx = Transaction(database, engine, mode=Mode.OPEN)
     rsp = tx.run(ctx, _list_edb_action())
     actions = rsp["actions"]
@@ -346,10 +433,10 @@ def list_edb(ctx: Context, database: str, engine: str) -> list:
     return rels
 
 
-# Returns a list of source names installed in the given database.
-def list_sources(ctx: Context, database: str, engine: str) -> list:
-    sources = _list_sources(ctx, database, engine)
-    return [source["name"] for source in sources]
+# Returns a list of models installed in the given database.
+def list_models(ctx: Context, database: str, engine: str) -> list:
+    models = _list_models(ctx, database, engine)
+    return [model["name"] for model in models]
 
 
 # Generate a rel literal relation for the given dict.
@@ -415,11 +502,11 @@ def _gen_syntax_config(syntax: dict = {}) -> str:
 #
 # Schema: a map from col name to rel type name, eg:
 #   {'a': "int", 'b': "string"}
-def load_csv(ctx: Context, database: str, engine: str, relation: str, data,
-             syntax: dict = {}) -> dict:
+def load_csv(ctx: Context, database: str, engine: str, relation: str,
+             data: str or io.TextIOBase, syntax: dict = {}) -> dict:
     if isinstance(data, str):
         pass  # ok
-    elif isinstance(data, io.TextIO):
+    elif isinstance(data, io.TextIOBase):
         data = data.read()
     else:
         raise TypeError(f"bad type for arg 'data': {data.__class__.__name__}")
@@ -430,10 +517,11 @@ def load_csv(ctx: Context, database: str, engine: str, relation: str, data,
     return query(ctx, database, engine, command, inputs=inputs, readonly=False)
 
 
-def load_json(ctx: Context, database: str, engine: str, relation: str, data) -> dict:
+def load_json(ctx: Context, database: str, engine: str, relation: str,
+              data: str or io.TextIOBase) -> dict:
     if isinstance(data, str):
         pass  # ok
-    elif isinstance(data, io.TextIO):
+    elif isinstance(data, io.TextIOBase):
         data = data.read()
     else:
         raise TypeError(f"bad type for arg 'data': {data.__class__.__name__}")
@@ -453,3 +541,7 @@ create_compute = create_engine  # deprecated, use create_engine
 delete_compute = delete_engine  # deprecated, use delete_engine
 get_compute = get_engine        # deprecated, use get_engine
 list_computes = list_engines    # deprecated, use list_engines
+list_edb = list_edbs            # deprecated, use list_edbs
+delete_source = delete_model    # deprecated, use delete_model
+get_source = get_model          # deprecated, use get_model
+list_sources = list_models      # deprecated, use list_models
